open! Core
open! Async_kernel
open! Import
open! Face

let%expect_test "[font_family_list]" =
  print_s [%sexp (font_family_list () : string list)];
  [%expect {| () |}];
  return ()
;;

let customization_group =
  Customization.Group.defgroup
    "test-group"
    [%here]
    ~docstring:"A test customization group"
    ~parents:[]
;;

let some_attribute_value (type a) (attribute : a Attribute.t) : a option =
  let open Attribute in
  match attribute with
  | Background -> Some (Color Color.red)
  | Box -> None
  | Font -> None
  | Font_family -> Some (Name "family")
  | Font_foundry -> Some (Name "foundry")
  | Foreground -> Some (Color Color.red)
  | Height -> Some (Scale_underlying_face 0.5)
  | Inherit -> None
  | Inverse_video -> Some Yes
  | Overline -> Some Foreground
  | Slant -> Some Italic
  | Stipple -> None
  | Strike_through -> Some Foreground
  | Underline -> Some Foreground
  | Weight -> Some Ultra_bold
  | Width -> Some Ultra_condensed
;;

let%expect_test "[is_relative]" =
  List.iter Attribute.Packed.all ~f:(fun (Attribute.Packed.T attribute) ->
    match some_attribute_value attribute with
    | None -> ()
    | Some value ->
      print_s
        [%message
          ""
            ~_:(T (attribute, value) : Attribute_and_value.t)
            ~is_relative:(Attribute.is_relative attribute value : bool)
            ~unspecified_is_relative:
              (Attribute.is_relative attribute (Attribute.unspecified_value attribute)
               : bool)]);
  [%expect
    {|
    ((Background (Color red))
     (is_relative             false)
     (unspecified_is_relative true))
    ((Font_family (Name family))
     (is_relative             false)
     (unspecified_is_relative true))
    ((Font_foundry (Name foundry))
     (is_relative             false)
     (unspecified_is_relative true))
    ((Foreground (Color red))
     (is_relative             false)
     (unspecified_is_relative true))
    ((Height (Scale_underlying_face 0.5))
     (is_relative             true)
     (unspecified_is_relative true))
    ((Inverse_video           Yes)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Overline                Foreground)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Slant                   Italic)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Strike_through          Foreground)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Underline               Foreground)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Weight                  Ultra_bold)
     (is_relative             false)
     (unspecified_is_relative true))
    ((Width                   Ultra_condensed)
     (is_relative             false)
     (unspecified_is_relative true)) |}];
  return ()
;;

let%expect_test "[merge]" =
  let test b1 b2 =
    print_s
      [%message
        ""
          ~_:(b1 : Background.t)
          ~_:(b2 : Background.t)
          ~merge:(Attribute.merge Background b1 b2 : Background.t)]
  in
  test (Color Color.red) Unspecified;
  [%expect {|
    ((Color red) Unspecified (merge (Color red))) |}];
  test (Color Color.red) (Color Color.red);
  [%expect {|
    ((Color red)
     (Color red)
     (merge (Color red))) |}];
  test (Color Color.red) (Color Color.black);
  [%expect {|
    ((Color red)
     (Color black)
     (merge (Color red))) |}];
  test (Color Color.black) (Color Color.red);
  [%expect {|
    ((Color black)
     (Color red)
     (merge (Color black))) |}];
  return ()
;;

let%expect_test "[attribute]" =
  let t = of_name "default" in
  List.iter Attribute.Packed.all ~f:(fun (Attribute.Packed.T attribute) ->
    print_s [%sexp (T (attribute, attribute_value t attribute) : Attribute_and_value.t)]);
  [%expect
    {|
    (Background Unspecified)
    (Box nil)
    (Font unspecified)
    (Font_family Unspecified)
    (Font_foundry Unspecified)
    (Foreground Unspecified)
    (Height Unspecified)
    (Inherit Unspecified)
    (Inverse_video No)
    (Overline Absent)
    (Slant Unspecified)
    (Stipple nil)
    (Strike_through Absent)
    (Underline Absent)
    (Weight Unspecified)
    (Width Unspecified) |}];
  return ()
;;

let%expect_test "[all_defined], [attributes]" =
  List.iter (all_defined ()) ~f:(fun t ->
    print_s [%message "" ~_:(t : t) ~_:(attributes t : Attribute_and_value.t list)]);
  [%expect
    {|
    (bold (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (bold-italic (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Italic)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (bookmark-menu-bookmark (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (bookmark-menu-heading (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (font-lock-type-face)))))
    (border (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (buffer-menu-buffer (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (button (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (link)))))
    (completions-annotations (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (italic)))))
    (completions-common-part (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (completions-first-difference (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (bold)))))
    (cursor (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (default (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Absent)
      (Overline       Absent)
      (Strike_through Absent)
      (Box            nil)
      (Inverse_video  No)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        nil)
      (Inherit        Unspecified)))
    (eldoc-highlight-function-argument (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (bold)))))
    (error (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (escape-glyph (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color brown))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (file-name-shadow (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (shadow)))))
    (fixed-pitch (
      (Font_family (Name Monospace))
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (fixed-pitch-serif (
      (Font_family (Name "Monospace Serif"))
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-builtin-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-comment-delimiter-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (font-lock-comment-face)))))
    (font-lock-comment-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Italic)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-constant-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Foreground)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-doc-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (font-lock-string-face)))))
    (font-lock-function-name-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-keyword-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-negation-char-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-preprocessor-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (font-lock-builtin-face)))))
    (font-lock-regexp-grouping-backslash (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (bold)))))
    (font-lock-regexp-grouping-construct (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (bold)))))
    (font-lock-string-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Italic)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-type-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Foreground)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-variable-name-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Italic)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (font-lock-warning-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (error)))))
    (fringe (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background (Color gray))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (glyphless-char (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (underline)))))
    (header-line (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Foreground)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  No)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (mode-line)))))
    (header-line-highlight (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (highlight)))))
    (help-argument-name (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (italic)))))
    (highlight (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (homoglyph (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color brown))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (internal-border (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (isearch (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (isearch-fail (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (italic (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Italic)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (lazy-highlight (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Foreground)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (line-number (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (shadow default)))))
    (line-number-current-line (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (line-number)))))
    (link (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (underline)))))
    (link-visited (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (link)))))
    (match (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background (Color gray))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (menu (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (minibuffer-prompt (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color "medium blue"))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (mode-line (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (mode-line-buffer-id (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (mode-line-emphasis (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (mode-line-highlight (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (highlight)))))
    (mode-line-inactive (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (mode-line)))))
    (mouse (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (mouse-drag-and-drop-region (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (region)))))
    (next-error (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (region)))))
    (nobreak-hyphen (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color brown))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (nobreak-space (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (query-replace (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (isearch)))))
    (read-multiple-choice-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (underline)))))
    (region (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background (Color gray))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (scroll-bar (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (secondary-selection (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (shadow (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (show-paren-match (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (underline)))))
    (show-paren-match-expression (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (show-paren-match)))))
    (show-paren-mismatch (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (success (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (tool-bar (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box (:line-width 1 :style released-button))
      (Inverse_video Unspecified)
      (Foreground (Color black))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (tooltip (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (variable-pitch)))))
    (trailing-whitespace (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Yes)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (tty-menu-disabled-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color white))
      (Background (Color blue))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (tty-menu-enabled-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color yellow))
      (Background (Color blue))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (tty-menu-selected-face (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background (Color red))
      (Stipple unspecified)
      (Inherit Unspecified)))
    (underline (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Foreground)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (variable-pitch (
      (Font_family (Name "Sans Serif"))
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (vc-conflict-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-edited-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-locally-added-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-locked-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-missing-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-needs-update-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-removed-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vc-state-base (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (vc-up-to-date-state (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (vc-state-base)))))
    (vertical-border (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit (Face (mode-line-inactive)))))
    (warning (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Bold)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground     Unspecified)
      (Background     Unspecified)
      (Stipple        unspecified)
      (Inherit        Unspecified)))
    (window-divider (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color gray60))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (window-divider-first-pixel (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color gray80))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified)))
    (window-divider-last-pixel (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Unspecified)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color gray40))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified))) |}];
  return ()
;;

let%expect_test "[defface]" =
  let name = "foobar" in
  let t =
    defface
      name
      [%here]
      ~docstring:"Test face"
      ~customization_group
      [ T (Foreground, Color Color.cyan); T (Overline, Absent) ]
  in
  print_s [%message "" ~_:(t : t) ~_:(attributes t : Attribute_and_value.t list)];
  [%expect
    {|
    (foobar (
      (Font_family    Unspecified)
      (Font_foundry   Unspecified)
      (Width          Unspecified)
      (Height         Unspecified)
      (Weight         Unspecified)
      (Slant          Unspecified)
      (Underline      Unspecified)
      (Overline       Absent)
      (Strike_through Unspecified)
      (Box            unspecified)
      (Inverse_video  Unspecified)
      (Foreground (Color cyan))
      (Background Unspecified)
      (Stipple    unspecified)
      (Inherit    Unspecified))) |}];
  return ()
;;
